#!/usr/bin/env python

"""Save the bunnies from the unseen menace.

Part of the "first demo game" from the original Gummworld. Reimplemented in
Gummworld2.

This file is in the Public Domain. No warranties, no restrictions.
"""

import sys
from random import randrange

import pygame
from pygame.locals import *

import paths
from gummworld2 import *

# This application
from bunny import Bunny
from bunny_collector import BunnyCollector


class App(Engine):
    
    def __init__(self, display_size):
        
        # Determine dimensions of the view areas: game and stats.
        display_size = Vec2d(display_size)
        margin = 30
        stats_height = 60
        view_width = display_size.x-margin*2
        # Game view.
        left,top = margin,margin
        height = display_size.y-margin*2-stats_height
        camera_view_rect = Rect(left, top, view_width, height)
        # Stats view.
        left,top = camera_view_rect.left, camera_view_rect.bottom
        stat_view_rect = Rect(left, top, view_width, stats_height)
        
        super(App, self).__init__(
            resolution=display_size,
            tile_size=(128,128), map_size=(20,20),
            camera_view_rect=camera_view_rect,
            frame_speed=0,
            world_type=QUADTREE_WORLD,
            world_args=dict(worst_case=99),
        )
        
        # Subsurfaces for game and stats display.
        self.game_screen = State.camera.view
        self.stat_screen = View(State.screen.surface, stat_view_rect)
        
        # Make the avatar.
        self.bunny_collector = BunnyCollector((128.0*2.9,128.0*1.9))
        State.camera.target = self.bunny_collector
        self.bunny_collector.loose = 25
        
        # Load the map (grass and trees).
        load_map()
        
        # Some sprite groups for management. And bunnies.
        self.tree_group = State.map.layers[1]
        self.bunny_group = BucketGroup(State.map.tile_size, State.map.map_size,
            *make_bunnies(self.bunny_collector.loose))
        
        # Populate the world with interactive objects.
        State.world.add(self.bunny_collector)
        State.world.add(*[s for s in self.tree_group if s])
        State.world.add(*self.bunny_group.sprites())
        
        # I like huds.
        toolkit.make_hud()
        State.show_hud = True
        
        # Make a font for stats.
        pygame_utils.make_font('Vera', data.filepath('font', 'Vera.ttf'), 16)
        
        # Hold cursor key-down states.
        self.move_x = 0
        self.move_y = 0
        
        State.camera.update()

    def update(self):
        # Do avatar movement and terrain collisions.
        self.update_camera_position()
        self.update_avatar_vs_trees()
        # Update camera.
        State.camera.update()
        # Do bunny updates and avatar collisions.
        self.update_bunnies()
        self.update_bunnies_vs_avatar()
        # Do stats and check status.
        State.hud.update()
        if self.bunny_collector.loose == 0:
            State.running = False
    
    def update_camera_position(self):
        """update the camera's position if any movement keys are held down
        """
        if not (self.move_y or self.move_x):
            return
        camera = State.camera
        wx,wy = camera.position + (self.move_x,self.move_y)
        rect = State.world.rect
        wx = max(min(wx,rect.right), rect.left)
        wy = max(min(wy,rect.bottom), rect.top)
        camera.position = wx,wy
        State.world.add(camera.target)
        
    def update_avatar_vs_trees(self):
        if not (self.move_y or self.move_x):
            return
        camera = State.camera
        target = camera.target
        tree_group = self.tree_group
        hits = []
        for s1,s2 in State.world.collisions:
            if s1 is target and s2 in tree_group:
                hits.append(s2)
        if hits:
            moved = camera.target_moved
            rect = camera.target.rect.copy()
            move_back = Vec2d(moved.x,0)
            rect.center = camera.target.rect.center - move_back
            ok = True
            for s in hits:
                if s.rect.colliderect(rect):
                    ok = False
                    break
            if ok:
                camera.position -= moved
                camera.position += (0,moved.y)
            else:
                ok = True
                move_back = Vec2d(0,moved.y)
                rect.center = camera.target.rect.center - move_back
                for s in hits:
                    if s.rect.colliderect(rect):
                        ok = False
                        break
                if ok:
                    camera.position -= moved
                    camera.position += (moved.x,0)
                else:
                    camera.position -= moved
                    camera.position += (0,0)
    
    def update_bunnies(self):
        self.bunny_group.update(State.camera.visible_tile_range[0])

    def update_bunnies_vs_avatar(self):
        target = State.camera.target
        bunny_group = self.bunny_group
        bunny_collector = self.bunny_collector
        for s1,s2 in State.world.collisions:
            if s1 is target and s2 in bunny_group:
                bunny_collector.hit(s2)
    
    def draw(self):
        State.camera.interpolate()
        
        State.screen.clear()
        
        toolkit.draw_tiles()
        self.bunny_group.draw(State.camera.visible_tile_range[0])
        self.draw_bunny_collector()
        self.draw_stats()
        State.hud.draw()
        
        main_surf = State.screen.surface
        pygame.draw.rect(
            main_surf, Color('grey'), self.game_screen.parent_rect, 1)
        pygame.draw.rect(
            main_surf, Color('grey'), self.stat_screen.parent_rect, 1)
        
        State.screen.flip()
    
    def draw_bunnies(self):
        blit = State.camera.surface.blit
        draw_sprite = toolkit.draw_sprite
        for s in self.bunnies_in_frame:
            draw_sprite(s)
    
    def draw_bunny_collector(self):
        game_screen = self.game_screen
        bunny_collector = self.bunny_collector
        game_screen.blit(
            bunny_collector.image,
            game_screen.center - bunny_collector.half_size)
    
    def draw_stats(self):
        surface = self.stat_screen.surface
        bunny_collector = self.bunny_collector
        draw_text(surface, 1, 'Bunnies collected: '+str(bunny_collector.collected))
        draw_text(surface, 2, 'Bunnies remaining: '+str(bunny_collector.loose))
    
    def on_key_down(self, unicode, key, mod):
        # Turn on key-presses.
        if key == K_DOWN:
            self.move_y += State.speed
        elif key == K_UP:
            self.move_y -= State.speed
        elif key == K_RIGHT:
            self.move_x += State.speed
        elif key == K_LEFT:
            self.move_x -= State.speed
        
    def on_key_up(self, key, mod):
        # Turn off key-presses.
        if key == K_DOWN:
            self.move_y -= State.speed
        elif key == K_UP:
            self.move_y += State.speed
        elif key == K_RIGHT:
            self.move_x -= State.speed
        elif key == K_LEFT:
            self.move_x += State.speed
        
    def on_quit(self):
        quit()


def load_map():
    # Load the grass in every tile.
    image,rect = pygame_utils.load_image(
        data.filepath('image', 'grass.bmp'))
    map = State.map
    map.layers.append(MapLayer(map.tile_size, map.map_size))
    for y in range(20):
        for x in range(20):
            s = pygame.sprite.Sprite()
            s.image = image
            s.rect = rect.copy()
            s.name = (x,y)
            s.rect.topleft = Vec2d(x,y) * map.tile_size
            map.add(s, layer=0)
    map.layers.append(MapLayer(map.tile_size, map.map_size))
    # Load a tree wherever there's a "T".
    tree_data = """\
TTTTTTTTTTTTTTTTTTTT
TTooooooooooooToooTT
TooTTTooTTToooTToooT
TooooooooToooooTTooT
ToooooooooooooooTToT
TooooooooooooooooooT
TooooooooooooooooooT
TooooooTTTooooooooTT
ToooooooooTooooooTTT
TooooooooooTToooooTT
TooooooooooTTooooooT
ToToToTooooooooooooT
TooooooooooooTTTToTT
TooToToToooooTooooTT
TooooooooooooTooTTTT
ToToToToTooooTooooTT
TooooooooooooTTTToTT
TooToToToTTooooooooT
TTooooooooooooooooTT
TTTTTTTTTTTTTTTTTTTT
""".split('\n')
    image,rect = pygame_utils.load_image(
        data.filepath('image', 'tree.bmp'), colorkey=-1)
    for y in range(20):
        for x in range(20):
            if tree_data[y][x] == 'T':
                s = pygame.sprite.Sprite()
                s.image = image
                s.rect = rect.copy()
                s.name = (x,y)
                s.rect.topleft = Vec2d(x,y) * map.tile_size
            else:
                s = None
            map.add(s, layer=1)


def make_bunnies(num):
    """Create num bunnies. Place them randomly on the map. They cannot sit in
    trees. Returns a list of bunny sprites.
    """
    
    # The return value.
    sprites = []
    
    # Speed-ups.
    map = State.map
    world_size = Vec2d(State.world.rect.size)
    
    # Make num bunnies...
    for i in range(num):
        
        # 1000 failures to place a bunny is a terminal error.
        stall = 1000
        
        bunny = Bunny()
        
        while 1:
            # Randomly place the sprite in world coordinates.
            bunny.position = randrange(world_size.x),randrange(world_size.y)
            
            # Convert position to map grid position, and screen position.
            mapx,mapy = bunny.position / map.tile_size
            mapx,mapy = int(round(mapx)),int(round(mapy))
            
            # For "this" tile and the eight adjacent tiles, check for collision
            # with any surrounding trees...
            bunny_placed = True
            for tile in map.get_tiles(mapx-1, mapy-1, mapx+1, mapy+1, layer=1):
                if tile and tile.rect.colliderect(bunny.rect):
                    bunny_placed = False
                    break
            if bunny_placed:
                break
            
            # Collision with a tree. Retry placement.
            stall -= 1
            if stall <= 0:
                print ' !!! 1000 failures for bunny placement: no unoccupied spaces on map?'
                raise pygame.error('Game initialization failed on count '+str(i))
        
        # Success, bunny placed.
        sprites.append(bunny)
    
    return sprites


def do_countdown(app):
    messages = ['Save the bunnies!', 'Ready...', 'Set...', 'GO!!!']
    game_screen = app.game_screen
    for nr in range(len(messages)):
        State.screen.clear()
        app.draw()
        draw_level_text(game_screen.surface, messages[nr])
        State.screen.flip()
        pygame.time.wait(1500)

def do_end_game(app):
    pygame.time.wait(1000)
    pygame.event.get()
    messages = ['Congratulations!!!', 'Title earned: Bunny Hugger']
    nr = 0
    game_screen = app.game_screen
    game_surface = game_screen.surface
    clear_copy = game_surface.copy()
    while 1:
        game_surface.blit(clear_copy, (0,0))
        draw_level_text(game_screen.surface, messages[nr])
        State.screen.flip()
        pygame.time.wait(1500)
        nr += 1
        if nr == len(messages):
            nr = 0
        for event in pygame.event.get():
            if event.type == KEYDOWN and event.key == K_ESCAPE:
                return


def draw_level_text(surface, text):
    font = pygame_utils.get_font('Vera')
    fimage = font.render(text, 1, (254,254,0))
    frect = fimage.get_rect()
    srect = surface.get_rect()
    frect.topleft = Vec2d(srect.center) - frect.center
    frect.top -= font.get_height()
    surface.blit(fimage, frect)


def draw_text(surface, row, text):
    font = pygame_utils.get_font('Vera')
    image = font.render(text, 1, Color('yellow'))
    rect = image.get_rect()
    rect.topleft = 5, 5 + (row-1) * font.get_height()
    surface.blit(image, rect)


def main():
    app = App((800,600))
    do_countdown(app)
    app.run()
    do_end_game(app)
